""" Constants """

DataType = {"CUSTOM": "CUSTOM"}

Operator = {
    "UNDEFINED": "UNDEFINED",
    "CONTAINS": "CONTAINS",
    "EXACT": "EXACT",
    "MATCH": "REGULAR_EXPRESSION",
    "LOWER": "LOWER",
    "EQUAL": "EQUAL",
    "GREATER": "GREATER",
    "IS_TRUE": "TRUE",
    "IS_FALSE": "FALSE",
    "AMONG_VALUES": "AMONG_VALUES",
    "ANY": "ANY",
}
